@@ -63,6 +63,8 @@ define([
         this._autorestart_attempt = 0;
         this._reconnect_attempt = 0;
         this.reconnect_limit = 7;
+        
+        this._pending_messages = [];
     };
 
     /**
@@ -274,7 +276,7 @@ define([
             }
         };
 
-        var url = utils.url_join_encode(this.kernel_url, 'interrupt');
+        var url = utils.url_path_join(this.kernel_url, 'interrupt');
         $.ajax(url, {
             processData: false,
             cache: false,
@@ -316,7 +318,7 @@ define([
             }
         };
 
-        var url = utils.url_join_encode(this.kernel_url, 'restart');
+        var url = utils.url_path_join(this.kernel_url, 'restart');
         $.ajax(url, {
             processData: false,
             cache: false,
@@ -337,7 +339,7 @@ define([
          * @function reconnect
          */
         if (this.is_connected()) {
-            return;
+            this.stop_channels();
         }
         this._reconnect_attempt = this._reconnect_attempt + 1;
         this.events.trigger('kernel_reconnecting.Kernel', {
@@ -362,7 +364,8 @@ define([
                 that.id = data.id;
                 that.name = data.name;
             }
-            that.kernel_url = utils.url_join_encode(that.kernel_service_url, that.id);
+            that.kernel_url = utils.url_path_join(that.kernel_service_url,
+                encodeURIComponent(that.id));
             if (success) {
                 success(data, status, xhr);
             }
@@ -394,7 +397,8 @@ define([
          * @param {Object} data - information about the kernel including id
          */
         this.id = data.id;
-        this.kernel_url = utils.url_join_encode(this.kernel_service_url, this.id);
+        this.kernel_url = utils.url_path_join(this.kernel_service_url,
+            encodeURIComponent(this.id));
         this.start_channels();
     };
 
@@ -407,6 +411,15 @@ define([
          * @function _kernel_connected
          */
         this.events.trigger('kernel_connected.Kernel', {kernel: this});
+
+        // Send pending messages. We shift the message off the queue
+        // after the message is sent so that if there is an exception,
+        // the message is still pending.
+        while (this._pending_messages.length > 0) {
+          this.ws.send(this._pending_messages[0]);
+          this._pending_messages.shift();
+        }
+
         // get kernel info so we know what state the kernel is in
         var that = this;
         this.kernel_info(function (reply) {
@@ -441,7 +454,7 @@ define([
         
         this.ws = new this.WebSocket([
                 that.ws_url,
-                utils.url_join_encode(that.kernel_url, 'channels'),
+                utils.url_path_join(that.kernel_url, 'channels'),
                 "?session_id=" + that.session_id
             ].join('')
         );
@@ -522,8 +535,13 @@ define([
 
         this.events.trigger('kernel_disconnected.Kernel', {kernel: this});
         if (error) {
-            console.log('WebSocket connection failed: ', ws_url);
-            this.events.trigger('kernel_connection_failed.Kernel', {kernel: this, ws_url: ws_url, attempt: this._reconnect_attempt});
+            console.log('WebSocket connection failed: ', ws_url, error);
+            this.events.trigger('kernel_connection_failed.Kernel', {
+                kernel: this,
+                ws_url: ws_url,
+                attempt: this._reconnect_attempt,
+                error: error,
+            });
         }
         this._schedule_reconnect();
     };
@@ -600,19 +618,34 @@ define([
         return (this.ws === null);
     };
     
+    Kernel.prototype._send = function(msg) {
+      /**
+       * Send a message (if the kernel is connected) or queue the message for future delivery
+       *
+       * Pending messages will automatically be sent when a kernel becomes connected.
+       *
+       * @function _send
+       * @param msg
+       */
+      if (this.is_connected()) {
+            this.ws.send(msg);
+        } else {
+            this._pending_messages.push(msg);
+        }
+    }
+    
     Kernel.prototype.send_shell_message = function (msg_type, content, callbacks, metadata, buffers) {
         /**
          * Send a message on the Kernel's shell channel
          *
+         * If the kernel is not connected, the message will be buffered.
+         * 
          * @function send_shell_message
          */
-        if (!this.is_connected()) {
-            throw new Error("kernel is not connected");
-        }
         var msg = this._get_msg(msg_type, content, metadata, buffers);
         msg.channel = 'shell';
-        this.ws.send(serialize.serialize(msg));
         this.set_callbacks_for_msg(msg.header.msg_id, callbacks);
+        this._send(serialize.serialize(msg));
         return msg.header.msg_id;
     };
 
@@ -625,7 +658,7 @@ define([
          *
          * When calling this method, pass a callback function that expects one argument.
          * The callback will be passed the complete `kernel_info_reply` message documented
-         * [here](http://ipython.org/ipython-doc/dev/development/messaging.html#kernel-info)
+         * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#kernel-info)
          */
         var callbacks;
         if (callback) {
@@ -634,13 +667,34 @@ define([
         return this.send_shell_message("kernel_info_request", {}, callbacks);
     };
 
+    Kernel.prototype.comm_info = function (target_name, callback) {
+        /**
+         * Get comm info
+         *
+         * @function comm_info
+         * @param callback {function}
+         *
+         * When calling this method, pass a callback function that expects one argument.
+         * The callback will be passed the complete `comm_info_reply` message documented
+         * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#comm_info)
+         */
+        var callbacks;
+        if (callback) {
+            callbacks = { shell : { reply : callback } };
+        }
+        var content = {
+            target_name : target_name,
+        };
+        return this.send_shell_message("comm_info_request", content, callbacks);
+    };
+
     Kernel.prototype.inspect = function (code, cursor_pos, callback) {
         /**
          * Get info on an object
          *
          * When calling this method, pass a callback function that expects one argument.
          * The callback will be passed the complete `inspect_reply` message documented
-         * [here](http://ipython.org/ipython-doc/dev/development/messaging.html#object-information)
+         * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#object-information)
          *
          * @function inspect
          * @param code {string}
@@ -731,7 +785,7 @@ define([
      * `complete_reply` message as its only argument when it arrives.
      *
      * `complete_reply` is documented
-     * [here](http://ipython.org/ipython-doc/dev/development/messaging.html#complete)
+     * [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#complete)
      *
      * @function complete
      * @param code {string}
@@ -754,16 +808,13 @@ define([
      * @function send_input_reply
      */
     Kernel.prototype.send_input_reply = function (input) {
-        if (!this.is_connected()) {
-            throw new Error("kernel is not connected");
-        }
         var content = {
             value : input
         };
         this.events.trigger('input_reply.Kernel', {kernel: this, content: content});
         var msg = this._get_msg("input_reply", content);
         msg.channel = 'stdin';
-        this.ws.send(serialize.serialize(msg));
+        this._send(serialize.serialize(msg));
         return msg.header.msg_id;
     };
 
@@ -862,20 +913,17 @@ define([
         this._msg_queue = this._msg_queue.then(function() {
             return serialize.deserialize(e.data);
         }).then(function(msg) {return that._finish_ws_message(msg);})
-        .catch(function(error) {console.error("Couldn't process kernel message", error) });
+        .catch(function(error) { console.error("Couldn't process kernel message", error); });
     };
 
     Kernel.prototype._finish_ws_message = function (msg) {
         switch (msg.channel) {
             case 'shell':
                 return this._handle_shell_reply(msg);
-                break;
             case 'iopub':
                 return this._handle_iopub_message(msg);
-                break;
             case 'stdin':
                 return this._handle_input_request(msg);
-                break;
             default:
                 console.error("unrecognized message channel", msg.channel, msg);
         }
@@ -898,7 +946,7 @@ define([
         this._finish_shell(parent_id);
         
         if (shell_callbacks.reply !== undefined) {
-            promise = promise.then(function() {return shell_callbacks.reply(reply)});
+            promise = promise.then(function() {return shell_callbacks.reply(reply);});
         }
         if (content.payload && shell_callbacks.payload) {
             promise = promise.then(function() {
